import numpy as np
from banpei.base.model import BaseModel
from .utils import *

class SST(BaseModel):
    def __init__(self, w, m=2, k=None, L=None):
        """
        Parameters
        ----------
        w    : int
               Window size
        m    : int
               Number of basis vectors
        k    : int
               Number of columns for the trajectory and test matrices
        L    : int
               Lag time
        """
        self.w = w
        self.m = m
        if k is None:
            self.k = self.w // 2
        else:
            self.k = k
        if L is None:
            self.L = self.k // 2
        else:
            self.L = L

    def detect(self, data, is_lanczos=False):
        """
        Batch mode detection

        Parameters
        ----------
        data : array_like
               Input array or object that can be converted to an array.
        is_lanczos : boolean
               If true, the change score is calculated based on the lanczos method

        Returns
        -------
            Numpy array contains the degree of change.
            The size of Numpy array is the same as input array.
        """
        # Set variables
        data = self.convert_to_nparray(data)
        T = len(data)

        # Check the size of input data
        if not len(data) > self.L + self.w + self.k - 2:
            raise ValueError("Input data is too small.")

        # Calculation range
        start_cal = self.k + self.w
        end_cal = T - self.L + 1

        # Calculate the degree of change
        change_scores = np.zeros(len(data))
        for t in range(start_cal, end_cal + 1):
            # Trajectory matrix
            start_tra = t - self.w - self.k + 1
            end_tra = t - self.w
            tra_matrix = self._extract_matrix(data, start_tra, end_tra, self.w)

            # Test matrix
            start_test = start_tra + self.L
            end_test = end_tra + self.L
            test_matrix = self._extract_matrix(data, start_test, end_test, self.w)

            # Calculate the score by singular value decomposition(SVD)
            if is_lanczos:
                change_scores[t] = self._calculate_score_by_lanczos(tra_matrix, test_matrix)
            else:
                change_scores[t] = self._calculate_score_by_svd(tra_matrix, test_matrix)

        return change_scores

    def stream_detect(self, data, is_lanczos=False):
        """
        Stream mode detection for live monitoring.

        Parameters
        ----------
        data : array_like
               Input array or object that can be converted to an array.
        is_lanczos : boolean
               If true, the change score is calculated based on the lanczos method

        Returns
        -------
        tuple: (score, delay)
              score means the degree of change in the latest we can calculate.
              delay means the time lag between the latest data point and calculation point.
        """
        # Set variables
        data = self.convert_to_nparray(data)
        T = len(data)

        # Check the size of input data
        if not len(data) > self.L + self.w + self.k - 2:
            return 0

        # Calculation range
        t = T - self.L + 1

        # Trajectory matrix
        start_tra = t - self.w - self.k + 1
        end_tra = t - self.w
        tra_matrix = self._extract_matrix(data, start_tra, end_tra, self.w)

        # Test matrix
        start_test = start_tra + self.L
        end_test = end_tra + self.L
        test_matrix = self._extract_matrix(data, start_test, end_test, self.w)

        # Calculate the change score
        if is_lanczos:
            return self._calculate_score_by_lanczos(tra_matrix, test_matrix)
        else:
            return self._calculate_score_by_svd(tra_matrix, test_matrix)

    def _calculate_score_by_svd(self, tra_matrix, test_matrix):
        U_tra, _, _ = np.linalg.svd(tra_matrix, full_matrices=False)
        U_test, _, _ = np.linalg.svd(test_matrix, full_matrices=False)
        U_tra_m = U_tra[:, :self.m]
        U_test_m = U_test[:, :self.m]
        s = np.linalg.svd(np.dot(U_tra_m.T, U_test_m), full_matrices=False, compute_uv=False)
        return 1 - s[0]

    def _calculate_score_by_lanczos(self, tra_matrix, test_matrix):
        m, _, _ = power_method(test_matrix)
        k = 2 * self.m if self.m % 2 == 0 else 2 * self.m - 1
        P = np.dot(tra_matrix, tra_matrix.T)
        T = tridiagonalize_by_lanczos(P, m, k)
        eigenvalue, eigenvectors = tridiag_eigen(T)
        return 1 - np.sum(eigenvectors[0, np.argsort(eigenvalue)[::-1][:self.m]] ** 2)

    def _extract_matrix(self, data, start, end, w):
        row = w
        column = end - start + 1
        matrix = np.empty((row, column))
        i = 0
        for t in range(start, end+1):
            matrix[:, i] = data[t-1:t-1+row]
            i += 1

        return matrix



import numpy as np


class BaseModel(object):
    def __init__(self):
        pass

    def convert_to_nparray(self, data):
        if not isinstance(data, np.ndarray):
            return np.array(data)
        else:
            return data

    def detect(self, *args):
        """
        This is a placeholder intended to be overwritten by individual models.
        """
        raise NotImplementedError



import numpy as np
from scipy import stats
from banpei.base.model import BaseModel


class Hotelling(BaseModel):
    def __init__(self):
        pass

    def detect(self, data, threshold):
        """
        Parameters
        ----------
        data : array_like
               Input array or object that can be converted to an array.
        threshold : float

        Returns
        -------
        List of tuples where each tuple contains index number and anomalous value.
        """
        data = self.convert_to_nparray(data)

        # Set the threshold of abnormality
        abn_th = stats.chi2.interval(1-threshold, 1)[1]

        # Covert raw data into the degree of abnormality
        avg = np.average(data)
        var = np.var(data)
        data_abn = [(x - avg)**2 / var for x in data]

        # Abnormality determination
        result = []
        for (index, x) in enumerate(data_abn):
            if x > abn_th:
                result.append((index, data[index]))
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['banpei', 'banpei.base']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.2,<2.0.0', 'pandas>=1.1.2,<2.0.0', 'scipy>=1.5.2,<2.0.0']

setup_kwargs = {
    'name': 'banpei',
    'version': '0.1.2',
    'description': 'Anomaly detection library with Python',
    'long_description': None,
    'author': 'Hirofumi Tsuruta',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import numpy as np


def power_method(A, iter_num=1):
    """
    Calculate the first singular vector/value of a target matrix based on the power method.
    Parameters
    ----------
    A : numpy array
        Target matrix
    iter_num : int
               Number of iterations

    Returns
    -------
    u : numpy array
        first left singular vector of A
    s : float
        first singular value of A
    v : numpy array
        first right singular vector of A
    """
    # set initial vector q
    q = np.random.normal(size=A.shape[1])
    q = q / np.linalg.norm(q)

    for i in range(iter_num):
        q = np.dot(np.dot(A.T, A), q)

    v = q / np.linalg.norm(q)
    Av = np.dot(A, v)
    s = np.linalg.norm(Av)
    u = Av / s

    return u, s, v

def tridiagonalize_by_lanczos(P, m, k):
    """
    Tridiagonalize matrix by lanczos method
    Parameters
    ----------
    P : numpy array
        Target matrix
    q : numpy array
        Initial vector
    k : int
        Size of the tridiagonal matrix

    Returns
    -------
    T : numpy array
        tridiagonal matrix
    """
    # Initialize variables
    T = np.zeros((k, k))
    r0 = m
    beta0 = 1
    q0 = np.zeros(m.shape)

    for i in range(k):
        q1 = r0 / beta0
        C = np.dot(P, q1)
        alpha1 = np.dot(q1, C)
        r1 = C - alpha1 * q1 - beta0 * q0
        beta1 = np.linalg.norm(r1)

        T[i, i] = alpha1
        if i + 1 < k:
          T[i, i + 1] = beta1
          T[i + 1, i] = beta1

        q0 = q1
        beta0 = beta1
        r0 = r1

    return T

def tridiag_eigen(T, iter_num=1, tol=1e-3):
    """
    Calculate eigenvalues and eigenvectors of tridiagonal matrix
    Parameters
    ----------
    P : numpy array
        Target matrix (tridiagonal)
    iter_num : int
        Number of iterations
    tol : float
        Stop iteration if the target matrix converges to a diagonal matrix with acceptable tolerance `tol`

    Returns
    -------
    eigenvalue : numpy array
                 Calculated eigenvalues
    eigenvectors : numpy array
                 Calculated eigenvectors
    """
    eigenvectors = np.identity(T.shape[0])

    for i in range(iter_num):
        Q, R = tridiag_qr_decomposition(T)
        T = np.dot(R, Q)
        eigenvectors = np.dot(eigenvectors, Q)
        eigenvalue = np.diag(T)
        if np.all((T - np.diag(eigenvalue) < tol)):
            break

    return eigenvalue, eigenvectors

def tridiag_qr_decomposition(T):
    """
    QR decomposition for a tridiagonal matrix
    Ref. http://www.ericmart.in/blog/optimizing_julia_tridiag_qr
    Parameters
    ----------
    T : numpy array
        Target matrix (tridiagonal)

    Returns
    -------
    Qt.T : numpy array
    R    : numpy array
    """
    R = T.copy()
    Qt = np.eye(T.shape[0])

    for i in range(T.shape[0] - 1):
        u = householder(R[i:i + 2, i])
        M = np.outer(u, u)
        R[i:i + 2, :(i + 3)] -= 2 * np.dot(M, R[i:i + 2, :(i + 3)])
        Qt[i:i + 2, :(i + 3)] -= 2 * np.dot(M, Qt[i:i + 2, :(i + 3)])

    return Qt.T, R

def householder(x):
    """
    Householder projection for vector.
    Parameters
    ----------
    x : numpy array
        Target vector

    Returns
    -------
    x : numpy array
    """
    x[0] = x[0] + np.sign(x[0]) * np.linalg.norm(x)
    x = x / np.linalg.norm(x)
    return x
